# -*- coding: utf-8 -*-

from typing import Optional, Callable
from brainpy import errors
from brainpy.base.base import Base
from brainpy.base import collector

math = None

__all__ = [
  'Function',
]


def _check_node(node):
  if not isinstance(node, Base):
    raise errors.BrainPyError(f'Element in "nodes" must be an instance of '
                              f'{Base.__name__}, but we got {type(node)}.')


def _check_var(var):
  global math
  if math is None: from brainpy import math
  if not isinstance(var, math.ndarray):
    raise errors.BrainPyError(f'Element in "dyn_vars" must be an instance of '
                              f'{math.ndarray.__name__}, but we got {type(var)}.')


class Function(Base):
  """The wrapper for Python functions.

  Parameters
  ----------
  f : function
    The function to wrap.
  nodes : optional, Base, sequence of Base, dict
    The nodes in the defined function ``f``.
  dyn_vars : optional, ndarray, sequence of ndarray, dict
    The dynamically changed variables.
  name : optional, str
    The function name.
  """

  def __init__(self, f: Optional[Callable], nodes=None, dyn_vars=None, name=None):
    # initialize 
    # ---
    self._f = f
    if name is None:
      name = self.unique_name(type_=f.__name__ if hasattr(f, '__name__') else 'Function')
    super(Function, self).__init__(name=name)

    # nodes 
    # ---
    if nodes is not None:
      self.implicit_nodes = collector.Collector()
      if isinstance(nodes, Base):
        nodes = (nodes,)
      if isinstance(nodes, (tuple, list)):
        for i, node in enumerate(nodes):
          _check_node(node)
          self.implicit_nodes[f'_node{i}'] = node
      elif isinstance(nodes, dict):
        for node in nodes.values():
          _check_node(node)
        self.implicit_nodes.update(nodes)
      else:
        raise ValueError(f'"nodes" only support list/tuple/dict of {Base.__name__}, '
                         f'but we got {type(nodes)}: {nodes}')

    # variables
    # ---
    if dyn_vars is not None:
      self.implicit_vars = collector.TensorCollector()
      global math
      if math is None: from brainpy import math
      if isinstance(dyn_vars, math.ndarray):
        dyn_vars = (dyn_vars,)
      if isinstance(dyn_vars, (tuple, list)):
        for i, v in enumerate(dyn_vars):
          _check_var(v)
          self.implicit_vars[f'_var{i}'] = v
      elif isinstance(dyn_vars, dict):
        for v in dyn_vars.values():
          _check_var(v)
        self.implicit_vars.update(dyn_vars)
      else:
        raise ValueError(f'"dyn_vars" only support list/tuple/dict of {math.ndarray.__name__}, '
                         f'but we got {type(dyn_vars)}: {dyn_vars}')

  def __call__(self, *args, **kwargs):
    return self._f(*args, **kwargs)



# -*- coding: utf-8 -*-


"""
This module provides methods to construct connectivity between neuron groups.
You can access them through ``brainpy.connect.XXX``.
"""

from .base import *
from .custom_conn import *
from .random_conn import *
from .regular_conn import *



# -*- coding: utf-8 -*-

import numpy as np

from brainpy.errors import ConnectorError

from .base import *

__all__ = [
  'FixedProb',
  'FixedPreNum',
  'FixedPostNum',
  'GaussianProb',

  'SmallWorld',
  'ScaleFreeBA',
  'ScaleFreeBADual',
  'PowerLaw',
]


# @tools.numba_jit
def _random_prob_conn(rng, pre_i, num_post, prob, include_self):
  p = rng.random(num_post) <= prob
  if (not include_self) and pre_i < num_post:
    p[pre_i] = False
  conn_j = np.asarray(np.where(p)[0], dtype=IDX_DTYPE)
  return conn_j


class FixedProb(TwoEndConnector):
  """Connect the post-synaptic neurons with fixed probability.

  Parameters
  ----------
  prob : float
      The conn probability.
  include_self : bool
      Whether create (i, i) conn?
  seed : optional, int
      Seed the random generator.
  """

  def __init__(self, prob, include_self=True, seed=None):
    super(FixedProb, self).__init__()
    assert 0. <= prob <= 1.
    self.prob = prob
    self.include_self = include_self
    self.seed = seed
    self.rng = np.random.RandomState(seed=seed)

  def build_conn(self):
    ind = []
    count = np.zeros(self.pre_num, dtype=IDX_DTYPE)

    for i in range(self.pre_num):
      posts = _random_prob_conn(self.rng, pre_i=i, num_post=self.post_num,
                                prob=self.prob, include_self=self.include_self)
      ind.append(posts)
      count[i] = len(posts)

    ind = np.concatenate(ind)
    indptr = np.concatenate(([0], count)).cumsum()

    return 'csr', (ind, indptr)


# @tools.numba_jit
def _fixed_num_prob(rng, num_need, num_total, i=0, include_self=False):
  prob = rng.random(num_total)
  if not include_self and i <= num_total:
    prob[i] = 1.
  neu_idx = np.argsort(prob)[:num_need]
  return np.asarray(neu_idx, dtype=IDX_DTYPE)


class FixedPreNum(TwoEndConnector):
  """Connect the pre-synaptic neurons with fixed number for each post-synaptic neuron.

  Parameters
  ----------
  num : float, int
    The connection probability (if "num" is float) or the fixed number of
    connectivity (if "num" is int).
  include_self : bool
    Whether create (i, i) conn ?
  seed : None, int
    Seed the random generator.

    - ``matrix``: This method will create a big matrix, then, the connectivity is constructed
      from this matrix :math:`(N_{pre}, N_{post})`. In a large network, this method will
      consume huge memories, including a matrix: :math:`(N_{pre}, N_{post})`, two vectors:
      :math:`2 * N_{need} * N_{post}`.
    - ``iter``: This method will iteratively build the synaptic connections. It has the
      minimum pressure of memory consuming, only :math:`2 * N_{need} * N_{post}`
      (``i`` and ``j`` vectors).
  """

  def __init__(self, num, include_self=True, seed=None):
    super(FixedPreNum, self).__init__()
    if isinstance(num, int):
      assert num >= 0, '"num" must be a non-negative integer.'
    elif isinstance(num, float):
      assert 0. <= num <= 1., '"num" must be in [0., 1.).'
    else:
      raise ConnectorError(f'Unknown type: {type(num)}')
    self.num = num
    self.seed = seed
    self.include_self = include_self
    self.rng = np.random.RandomState(seed=seed)

  def build_conn(self):
    # check
    if isinstance(self.num, int):
      assert 0 <= self.num <= self.pre_num, f'"num" must be smaller than "self.pre_num", ' \
                                            f'but got {self.num} > {self.pre_num}'
      num = self.num
    else:
      assert 0. <= self.num <= 1., f'"num" must be in [0., 1.), but got {self.num}'
      num = int(self.pre_num * self.num)

    pre_ids = []
    for i in range(self.post_num):
      pres = _fixed_num_prob(rng=self.rng, num_need=num, num_total=self.pre_num,
                             i=i, include_self=self.include_self)
      pre_ids.append(pres)
    pre_ids = np.concatenate(pre_ids)
    post_ids = np.repeat(np.arange(self.post_num), num)

    return 'ij', (pre_ids, post_ids)


class FixedPostNum(TwoEndConnector):
  """Connect the post-synaptic neurons with fixed number for each pre-synaptic neuron.

  Parameters
  ----------
  num : float, int
      The conn probability (if "num" is float) or the fixed number of
      connectivity (if "num" is int).
  include_self : bool
      Whether create (i, i) conn ?
  seed : None, int
      Seed the random generator.
  method : str
    The method used to create the connection.

    - ``matrix``: This method will create a big matrix, then, the connectivity is constructed
      from this matrix :math:`(N_{pre}, N_{post})`. In a large network, this method will
      consume huge memories, including a matrix: :math:`(N_{pre}, N_{post})`, two vectors:
      :math:`2 * N_{need} * N_{pre}`.
    - ``iter``: This method will iteratively build the synaptic connections. It has the
      minimum pressure of memory consuming, only :math:`2 * N_{need} * N_{pre}`
      (``i`` and ``j`` vectors).
  """

  def __init__(self, num, include_self=True, seed=None):
    super(FixedPostNum, self).__init__()
    if isinstance(num, int):
      assert num >= 0, '"num" must be a non-negative integer.'
    elif isinstance(num, float):
      assert 0. <= num <= 1., '"num" must be in [0., 1.).'
    else:
      raise ConnectorError(f'Unknown type: {type(num)}')
    self.num = num
    self.seed = seed
    self.include_self = include_self
    self.rng = np.random.RandomState(seed=seed)

  def build_conn(self):
    # check
    if isinstance(self.num, int):
      assert 0 <= self.num <= self.post_num, f'"num" must be smaller than "self.post_num", ' \
                                             f'but got {self.num} > {self.post_num}'
      prob = self.num / self.post_num
      num = self.num
    else:
      assert 0. <= self.num <= 1., f'"num" must be in [0., 1.), but got {self.num}'
      num = int(self.post_num * self.num)
      prob = self.num

    post_ids = []  # i.e. post_ids
    for i in range(self.pre_num):
      posts = _fixed_num_prob(rng=self.rng, num_need=num, num_total=self.post_num,
                              i=i, include_self=self.include_self)
      post_ids.append(posts)

    post_ids = np.concatenate(post_ids)
    count = np.ones(self.pre_num, dtype=IDX_DTYPE) * num
    indptr = np.concatenate(([0], count)).cumsum()

    return 'csr', (post_ids, indptr)


class GaussianProb(OneEndConnector):
  r"""Builds a Gaussian connectivity pattern within a population of neurons,
  where the connection probability decay according to the gaussian function.

  Specifically, for any pair of neurons :math:`(i, j)`,

  .. math::

      p(i, j)=\exp(-\frac{\sum_{k=1}^n |v_k^i - v_k^j|^2 }{2\sigma^2})

  where :math:`v_k^i` is the $i$-th neuron's encoded value at dimension $k$.

  Parameters
  ----------
  sigma : float
      Width of the Gaussian function.
  encoding_values : optional, list, tuple, int, float
    The value ranges to encode for neurons at each axis.

    - If `values` is not provided, the neuron only encodes each positional
      information, i.e., :math:`(i, j, k, ...)`, where :math:`i, j, k` is
      the index in the high-dimensional space.
    - If `values` is a single tuple/list of int/float, neurons at each dimension
      will encode the same range of values. For example, `values=(0, np.pi)`,
      neurons at each dimension will encode a continuous value space `[0, np.pi]`.
    - If `values` is a tuple/list of list/tuple, it means the value space will be
      different for each dimension. For example, `values=((-np.pi, np.pi), (10, 20), (0, 2 * np.pi))`.

  periodic_boundary : bool
    Whether the neuron encode the value space with the periodic boundary.
  normalize : bool
      Whether normalize the connection probability .
  include_self : bool
      Whether create the conn at the same position.
  seed : int
      The random seed.
  """

  def __init__(self, sigma, encoding_values=None, normalize=True, include_self=True,
               periodic_boundary=False, seed=None):
    super(GaussianProb, self).__init__()
    self.sigma = sigma
    self.encoding_values = encoding_values
    self.normalize = normalize
    self.include_self = include_self
    self.periodic_boundary = periodic_boundary
    self.seed = seed
    self.rng = np.random.RandomState(seed)

  def build_conn(self):
    # value range to encode
    if self.encoding_values is None:
      value_ranges = tuple([(0, s) for s in self.pre_size])
    elif isinstance(self.encoding_values, (tuple, list)):
      if len(self.encoding_values) == 0:
        raise ConnectorError(f'encoding_values has a length of 0.')
      elif isinstance(self.encoding_values[0], (int, float)):
        assert len(self.encoding_values) == 2
        assert self.encoding_values[0] < self.encoding_values[1]
        value_ranges = tuple([self.encoding_values for _ in self.pre_size])
      elif isinstance(self.encoding_values[0], (tuple, list)):
        if len(self.encoding_values) != len(self.pre_size):
          raise ConnectorError(f'The network size has {len(self.pre_size)} dimensions, while '
                               f'the encoded values provided only has {len(self.encoding_values)}-D. '
                               f'Error in {str(self)}.')
        for v in self.encoding_values:
          assert isinstance(v[0], (int, float))
          assert len(v) == 2
        value_ranges = tuple(self.encoding_values)
      else:
        raise ConnectorError(f'Unsupported encoding values: {self.encoding_values}')
    else:
      raise ConnectorError(f'Unsupported encoding values: {self.encoding_values}')

    # values
    values = [np.linspace(vs[0], vs[1], n + 1)[:n] for vs, n in zip(value_ranges, self.pre_size)]
    post_values = np.stack([v.flatten() for v in np.meshgrid(*values)])
    value_sizes = np.array([v[1] - v[0] for v in value_ranges])
    if value_sizes.ndim < post_values.ndim:
      value_sizes = np.expand_dims(value_sizes, axis=tuple([i + 1 for i in range(post_values.ndim - 1)]))

    # probability of connections
    prob_mat = []
    for i in range(self.pre_num):
      # values for node i
      i_coordinate = tuple()
      for s in self.pre_size[:-1]:
        i, pos = divmod(i, s)
        i_coordinate += (pos,)
      i_coordinate += (i,)
      i_value = np.array([values[i][c] for i, c in enumerate(i_coordinate)])
      if i_value.ndim < post_values.ndim:
        i_value = np.expand_dims(i_value, axis=tuple([i + 1 for i in range(post_values.ndim - 1)]))
      # distances
      dists = np.abs(i_value - post_values)
      if self.periodic_boundary:
        dists = np.where(dists > value_sizes / 2, value_sizes - dists, dists)
      exp_dists = np.exp(-(np.linalg.norm(dists, axis=0) / self.sigma) ** 2 / 2)
      prob_mat.append(exp_dists)
    prob_mat = np.stack(prob_mat)
    if self.normalize:
      prob_mat /= prob_mat.max()

    # connectivity
    conn_mat = prob_mat >= self.rng.random(prob_mat.shape)

    if not self.include_self:
      np.fill_diagonal(conn_mat, False)

    return 'mat', conn_mat


# @tools.numba_jit
def _smallworld_rewire(prob, i, all_j, include_self):
  if np.random.random(1) < prob:
    non_connected = np.where(all_j == False)[0]
    if len(non_connected) <= 1:
      return -1
    # Enforce no self-loops or multiple edges
    w = np.random.choice(non_connected)
    while (not include_self) and w == i:
      # non_connected.remove(w)
      w = np.random.choice(non_connected)
    return w
  else:
    return -1


class SmallWorld(TwoEndConnector):
  """Build a Watts–Strogatz small-world graph.

  Parameters
  ----------
  num_neighbor : int
      Each node is joined with its `k` nearest neighbors in a ring
      topology.
  prob : float
      The probability of rewiring each edge
  directed : bool
      Whether the graph is a directed graph.
  include_self : bool
      Whether include the node self.

  Notes
  -----
  First create a ring over :math:`num\_node` nodes [1]_.  Then each node in the ring is
  joined to its :math:`num\_neighbor` nearest neighbors (or :math:`num\_neighbor - 1` neighbors
  if :math:`num\_neighbor` is odd). Then shortcuts are created by replacing some edges as
  follows: for each edge :math:`(u, v)` in the underlying ":math:`num\_node`-ring with
  :math:`num\_neighbor` nearest neighbors" with probability :math:`prob` replace it with a new
  edge :math:`(u, w)` with uniformly random choice of existing node :math:`w`.

  References
  ----------
  .. [1] Duncan J. Watts and Steven H. Strogatz,
         Collective dynamics of small-world networks,
         Nature, 393, pp. 440--442, 1998.
  """

  def __init__(self, num_neighbor, prob, directed=False, include_self=False):
    super(SmallWorld, self).__init__()
    self.prob = prob
    self.directed = directed
    self.num_neighbor = num_neighbor
    self.include_self = include_self

  def build_conn(self):
    assert self.pre_size == self.post_size

    if isinstance(self.pre_size, int) or (isinstance(self.pre_size, (tuple, list)) and len(self.pre_size) == 1):
      num_node = self.pre_num

      if self.num_neighbor > num_node:
        raise ConnectorError("num_neighbor > num_node, choose smaller num_neighbor or larger num_node")
      # If k == n, the graph is complete not Watts-Strogatz
      if self.num_neighbor == num_node:
        conn = np.ones((num_node, num_node), dtype=MAT_DTYPE)
      else:
        conn = np.zeros((num_node, num_node), dtype=MAT_DTYPE)
        nodes = np.array(list(range(num_node)))  # nodes are labeled 0 to n-1
        # connect each node to k/2 neighbors
        for j in range(1, self.num_neighbor // 2 + 1):
          targets = np.concatenate([nodes[j:], nodes[0:j]])  # first j nodes are now last in list
          conn[nodes, targets] = True
          conn[targets, nodes] = True

        # rewire edges from each node
        # loop over all nodes in order (label) and neighbors in order (distance)
        # no self loops or multiple edges allowed
        for j in range(1, self.num_neighbor // 2 + 1):  # outer loop is neighbors
          targets = np.concatenate([nodes[j:], nodes[0:j]])  # first j nodes are now last in list
          if self.directed:
            # inner loop in node order
            for u, v in zip(nodes, targets):
              w = _smallworld_rewire(prob=self.prob, i=u, all_j=conn[u], include_self=self.include_self)
              if w != -1:
                conn[u, v] = False
                conn[u, w] = True
              w = _smallworld_rewire(prob=self.prob, i=u, all_j=conn[:, u], include_self=self.include_self)
              if w != -1:
                conn[v, u] = False
                conn[w, u] = True
          else:
            # inner loop in node order
            for u, v in zip(nodes, targets):
              w = _smallworld_rewire(prob=self.prob, i=u, all_j=conn[u], include_self=self.include_self)
              if w != -1:
                conn[u, v] = False
                conn[v, u] = False
                conn[u, w] = True
                conn[w, u] = True
        # conn = np.asarray(conn, dtype=MAT_DTYPE)
    else:
      raise ConnectorError('Currently only support 1D ring connection.')

    return 'mat', conn


def _random_subset(seq, m, rng):
  """Return m unique elements from seq.

  This differs from random.sample which can return repeated
  elements if seq holds repeated elements.

  Note: rng is a random.Random or numpy.random.RandomState instance.
  """
  targets = set()
  while len(targets) < m:
    x = rng.choice(seq)
    targets.add(x)
  return targets


class ScaleFreeBA(TwoEndConnector):
  """Build a random graph according to the Barabási–Albert preferential
  attachment model.

  A graph of :math:`num\_node` nodes is grown by attaching new nodes each with
  :math:`m` edges that are preferentially attached to existing nodes
  with high degree.

  Parameters
  ----------
  m : int
      Number of edges to attach from a new node to existing nodes
  seed : integer, random_state, or None (default)
      Indicator of random number generation state.

  Raises
  ------
  ConnectorError
      If `m` does not satisfy ``1 <= m < n``.

  References
  ----------
  .. [1] A. L. Barabási and R. Albert "Emergence of scaling in
         random networks", Science 286, pp 509-512, 1999.
  """

  def __init__(self, m, directed=False, seed=None):
    super(ScaleFreeBA, self).__init__()
    self.m = m
    self.directed = directed
    self.seed = seed
    self.rng = np.random.RandomState(seed)

  def build_conn(self):
    assert self.pre_num == self.post_num
    num_node = self.pre_num
    if self.m < 1 or self.m >= num_node:
      raise ConnectorError(f"Barabási–Albert network must have m >= 1 and "
                           f"m < n, while m = {self.m} and n = {num_node}")

    # Add m initial nodes (m0 in barabasi-speak)
    conn = np.zeros((num_node, num_node), dtype=MAT_DTYPE)
    # Target nodes for new edges
    targets = list(range(self.m))
    # List of existing nodes, with nodes repeated once for each adjacent edge
    repeated_nodes = []
    # Start adding the other n-m nodes. The first node is m.
    source = self.m
    while source < num_node:
      # Add edges to m nodes from the source.
      origins = [source] * self.m
      conn[origins, targets] = True
      if not self.directed:
        conn[targets, origins] = True
      # Add one node to the list for each new edge just created.
      repeated_nodes.extend(targets)
      # And the new node "source" has m edges to add to the list.
      repeated_nodes.extend([source] * self.m)
      # Now choose m unique nodes from the existing nodes
      # Pick uniformly from repeated_nodes (preferential attachment)
      targets = list(_random_subset(repeated_nodes, self.m, self.rng))
      source += 1

    return 'mat', conn


class ScaleFreeBADual(TwoEndConnector):
  r"""Build a random graph according to the dual Barabási–Albert preferential
  attachment model.

  A graph of :math::`num\_node` nodes is grown by attaching new nodes each with either $m_1$
  edges (with probability :math:`p`) or :math:`m_2` edges (with probability :math:`1-p`) that
  are preferentially attached to existing nodes with high degree.

  Parameters
  ----------
  m1 : int
      Number of edges to attach from a new node to existing nodes with probability :math:`p`
  m2 : int
      Number of edges to attach from a new node to existing nodes with probability :math:`1-p`
  p : float
      The probability of attaching :math:`m\_1` edges (as opposed to :math:`m\_2` edges)
  seed : integer, random_state, or None (default)
      Indicator of random number generation state.

  Raises
  ------
  ConnectorError
      If `m1` and `m2` do not satisfy ``1 <= m1,m2 < n`` or `p` does not satisfy ``0 <= p <= 1``.

  References
  ----------
  .. [1] N. Moshiri "The dual-Barabasi-Albert model", arXiv:1810.10538.
  """

  def __init__(self, m1, m2, p, directed=False, seed=None):
    super(ScaleFreeBADual, self).__init__()
    self.m1 = m1
    self.m2 = m2
    self.p = p
    self.directed = directed
    self.seed = seed
    self.rng = np.random.RandomState(seed=seed)

  def build_conn(self):
    assert self.pre_num == self.post_num
    num_node = self.pre_num
    if self.m1 < 1 or self.m1 >= num_node:
      raise ConnectorError(f"Dual Barabási–Albert network must have m1 >= 1 and m1 < num_node, "
                           f"while m1 = {self.m1} and num_node = {num_node}.")
    if self.m2 < 1 or self.m2 >= num_node:
      raise ConnectorError(f"Dual Barabási–Albert network must have m2 >= 1 and m2 < num_node, "
                           f"while m2 = {self.m2} and num_node = {num_node}.")
    if self.p < 0 or self.p > 1:
      raise ConnectorError(f"Dual Barabási–Albert network must have 0 <= p <= 1, while p = {self.p}")

    # Add max(m1,m2) initial nodes (m0 in barabasi-speak)
    conn = np.zeros((num_node, num_node), dtype=MAT_DTYPE)
    # List of existing nodes, with nodes repeated once for each adjacent edge
    repeated_nodes = []
    # Start adding the remaining nodes.
    source = max(self.m1, self.m2)
    # Pick which m to use first time (m1 or m2)
    m = self.m1 if self.rng.random() < self.p else self.m2
    # Target nodes for new edges
    targets = list(range(m))
    while source < num_node:
      # Add edges to m nodes from the source.
      origins = [source] * m
      conn[origins, targets] = True
      if not self.directed:
        conn[targets, origins] = True
      # Add one node to the list for each new edge just created.
      repeated_nodes.extend(targets)
      # And the new node "source" has m edges to add to the list.
      repeated_nodes.extend([source] * m)
      # Pick which m to use next time (m1 or m2)
      m = self.m1 if self.rng.random() < self.p else self.m2
      # Now choose m unique nodes from the existing nodes
      # Pick uniformly from repeated_nodes (preferential attachment)
      targets = list(_random_subset(repeated_nodes, m, self.rng))
      source += 1

    return 'mat', conn


class PowerLaw(TwoEndConnector):
  """Holme and Kim algorithm for growing graphs with powerlaw
  degree distribution and approximate average clustering.

  Parameters
  ----------
  m : int
      the number of random edges to add for each new node
  p : float,
      Probability of adding a triangle after adding a random edge
  seed : integer, random_state, or None (default)
      Indicator of random number generation state.

  Notes
  -----
  The average clustering has a hard time getting above a certain
  cutoff that depends on :math:`m`.  This cutoff is often quite low.  The
  transitivity (fraction of triangles to possible triangles) seems to
  decrease with network size.

  It is essentially the Barabási–Albert (BA) growth model with an
  extra step that each random edge is followed by a chance of
  making an edge to one of its neighbors too (and thus a triangle).

  This algorithm improves on BA in the sense that it enables a
  higher average clustering to be attained if desired.

  It seems possible to have a disconnected graph with this algorithm
  since the initial :math:`m` nodes may not be all linked to a new node
  on the first iteration like the BA model.

  Raises
  ------
  ConnectorError
      If :math:`m` does not satisfy :math:`1 <= m <= n` or :math:`p` does not
      satisfy :math:`0 <= p <= 1`.

  References
  ----------
  .. [1] P. Holme and B. J. Kim,
         "Growing scale-free networks with tunable clustering",
         Phys. Rev. E, 65, 026107, 2002.
  """

  def __init__(self, m, p, directed=False, seed=None):
    super(PowerLaw, self).__init__()
    self.m = m
    self.p = p
    if self.p > 1 or self.p < 0:
      raise ConnectorError(f"p must be in [0,1], while p={self.p}")
    self.directed = directed
    self.seed = seed
    self.rng = np.random.RandomState(seed)

  def build_conn(self):
    assert self.pre_num == self.post_num
    num_node = self.pre_num
    if self.m < 1 or num_node < self.m:
      raise ConnectorError(f"Must have m>1 and m<n, while m={self.m} and n={num_node}")
    # add m initial nodes (m0 in barabasi-speak)
    conn = np.zeros((num_node, num_node), dtype=MAT_DTYPE)
    repeated_nodes = list(range(self.m))  # list of existing nodes to sample from
    # with nodes repeated once for each adjacent edge
    source = self.m  # next node is m
    while source < num_node:  # Now add the other n-1 nodes
      possible_targets = _random_subset(repeated_nodes, self.m, self.rng)
      # do one preferential attachment for new node
      target = possible_targets.pop()
      conn[source, target] = True
      if not self.directed:
        conn[target, source] = True
      repeated_nodes.append(target)  # add one node to list for each new link
      count = 1
      while count < self.m:  # add m-1 more new links
        if self.rng.random() < self.p:  # clustering step: add triangle
          neighbors = np.where(conn[target])[0]
          neighborhood = [nbr for nbr in neighbors if not conn[source, nbr] and not nbr == source]
          if neighborhood:  # if there is a neighbor without a link
            nbr = self.rng.choice(neighborhood)
            conn[source, nbr] = True  # add triangle
            if not self.directed:
              conn[nbr, source] = True
            repeated_nodes.append(nbr)
            count = count + 1
            continue  # go to top of while loop
        # else do preferential attachment step if above fails
        target = possible_targets.pop()
        conn[source, target] = True
        if not self.directed:
          conn[target, source] = True
        repeated_nodes.append(target)
        count = count + 1
      repeated_nodes.extend([source] * self.m)  # add source node to list m times
      source += 1

    return 'mat', conn




# -*- coding: utf-8 -*-

import numpy as np

from brainpy import math as bm
from brainpy.tools.others import to_size, size2num
from .base import IntraLayerInitializer


__all__ = [
  'GaussianDecay',
  'DOGDecay',
]


class GaussianDecay(IntraLayerInitializer):
  r"""Builds a Gaussian connectivity pattern within a population of neurons,
  where the weights decay with gaussian function.

  Specifically, for any pair of neurons :math:`(i, j)`, the weight is computed as

  .. math::

      w(i, j) = w_{max} \cdot \exp(-\frac{\sum_{k=1}^n |v_k^i - v_k^j|^2 }{2\sigma^2})

  where :math:`v_k^i` is the $i$-th neuron's encoded value at dimension $k$.

  Parameters
  ----------
  sigma : float
    Width of the Gaussian function.
  max_w : float
    The weight amplitude of the Gaussian function.
  min_w : float, None
    The minimum weight value below which synapses are not created (default: :math:`0.005 * max\_w`).
  include_self : bool
    Whether create the conn at the same position.
  encoding_values : optional, list, tuple, int, float
    The value ranges to encode for neurons at each axis.

    - If `values` is not provided, the neuron only encodes each positional
      information, i.e., :math:`(i, j, k, ...)`, where :math:`i, j, k` is
      the index in the high-dimensional space.
    - If `values` is a single tuple/list of int/float, neurons at each dimension
      will encode the same range of values. For example, `values=(0, np.pi)`,
      neurons at each dimension will encode a continuous value space `[0, np.pi]`.
    - If `values` is a tuple/list of list/tuple, it means the value space will be
      different for each dimension. For example, `values=((-np.pi, np.pi), (10, 20), (0, 2 * np.pi))`.
  periodic_boundary : bool
    Whether the neuron encode the value space with the periodic boundary.
  normalize : bool
      Whether normalize the connection probability.
  """

  def __init__(self, sigma, max_w, min_w=None, encoding_values=None,
               periodic_boundary=False, include_self=True, normalize=False):
    super(GaussianDecay, self).__init__()
    self.sigma = sigma
    self.max_w = max_w
    self.min_w = max_w * 0.005 if min_w is None else min_w
    self.encoding_values = encoding_values
    self.periodic_boundary = periodic_boundary
    self.include_self = include_self
    self.normalize = normalize

  def __call__(self, shape, dtype=None):
    """Build the weights.

    Parameters
    ----------
    shape : tuple of int, list of int, int
      The network shape. Note, this is not the weight shape.
    """
    shape = to_size(shape)
    net_size = size2num(shape)

    # value ranges to encode
    if self.encoding_values is None:
      value_ranges = tuple([(0, s) for s in shape])
    elif isinstance(self.encoding_values, (tuple, list)):
      if len(self.encoding_values) == 0:
        raise ValueError
      elif isinstance(self.encoding_values[0], (int, float)):
        assert len(self.encoding_values) == 2
        assert self.encoding_values[0] < self.encoding_values[1]
        value_ranges = tuple([self.encoding_values for _ in shape])
      elif isinstance(self.encoding_values[0], (tuple, list)):
        if len(self.encoding_values) != len(shape):
          raise ValueError(f'The network size has {len(shape)} dimensions, while '
                           f'the encoded values provided only has {len(self.encoding_values)}-D. '
                           f'Error in {str(self)}.')
        for v in self.encoding_values:
          assert isinstance(v[0], (int, float))
          assert len(v) == 2
        value_ranges = tuple(self.encoding_values)
      else:
        raise ValueError(f'Unsupported encoding values: {self.encoding_values}')
    else:
      raise ValueError(f'Unsupported encoding values: {self.encoding_values}')

    # values
    values = [np.linspace(vs[0], vs[1], n + 1)[:n] for vs, n in zip(value_ranges, shape)]
    post_values = np.stack([v.flatten() for v in np.meshgrid(*values)])
    value_sizes = np.array([v[1] - v[0] for v in value_ranges])
    if value_sizes.ndim < post_values.ndim:
      value_sizes = np.expand_dims(value_sizes, axis=tuple([i + 1 for i in range(post_values.ndim - 1)]))

    # connectivity matrix
    conn_mat = []
    for i in range(net_size):
      # values for node i
      i_coordinate = tuple()
      for s in shape[:-1]:
        i, pos = divmod(i, s)
        i_coordinate += (pos,)
      i_coordinate += (i,)
      i_value = np.array([values[i][c] for i, c in enumerate(i_coordinate)])
      if i_value.ndim < post_values.ndim:
        i_value = np.expand_dims(i_value, axis=tuple([i + 1 for i in range(post_values.ndim - 1)]))
      # distances
      dists = np.abs(i_value - post_values)
      if self.periodic_boundary:
        dists = np.where(dists > value_sizes / 2, value_sizes - dists, dists)
      exp_dists = np.exp(-(np.linalg.norm(dists, axis=0) / self.sigma) ** 2 / 2)
      conn_mat.append(exp_dists)
    conn_mat = np.stack(conn_mat)
    if self.normalize:
      conn_mat /= conn_mat.max()
    if not self.include_self:
      np.fill_diagonal(conn_mat, 0.)

    # connectivity weights
    conn_weights = conn_mat * self.max_w
    conn_weights = np.where(conn_weights < self.min_w, 0., conn_weights)
    return bm.asarray(conn_weights, dtype=dtype)

  def __repr__(self):
    name = self.__class__.__name__
    bank = ' ' * len(name)
    return (f'{name}(sigma={self.sigma}, max_w={self.max_w}, min_w={self.min_w}, \n'
            f'{bank}periodic_boundary={self.periodic_boundary}, '
            f'include_self={self.include_self}, '
            f'normalize={self.normalize})')


class DOGDecay(IntraLayerInitializer):
  r"""Builds a Difference-Of-Gaussian (dog) connectivity pattern within a population of neurons.

  Mathematically, for the given pair of neurons :math:`(i, j)`, the weight between them is computed as

  .. math::

      w(i, j) = w_{max}^+ \cdot \exp(-\frac{\sum_{k=1}^n |v_k^i - v_k^j|^2}{2\sigma_+^2}) -
                w_{max}^- \cdot \exp(-\frac{\sum_{k=1}^n |v_k^i - v_k^j|^2}{2\sigma_-^2})

  where weights smaller than :math:`0.005 * max(w_{max}, w_{min})` are not created and
  self-connections are avoided by default (parameter allow_self_connections).

  Parameters
  ----------
  sigmas : tuple
      Widths of the positive and negative Gaussian functions.
  max_ws : tuple
      The weight amplitudes of the positive and negative Gaussian functions.
  min_w : float, None
      The minimum weight value below which synapses are not created (default: :math:`0.005 * min(max\_ws)`).
  include_self : bool
    Whether create the connections at the same position (self-connections).
  normalize : bool
    Whether normalize the connection probability .
  encoding_values : optional, list, tuple, int, float
    The value ranges to encode for neurons at each axis.

    - If `values` is not provided, the neuron only encodes each positional
      information, i.e., :math:`(i, j, k, ...)`, where :math:`i, j, k` is
      the index in the high-dimensional space.
    - If `values` is a single tuple/list of int/float, neurons at each dimension
      will encode the same range of values. For example, `values=(0, np.pi)`,
      neurons at each dimension will encode a continuous value space `[0, np.pi]`.
    - If `values` is a tuple/list of list/tuple, it means the value space will be
      different for each dimension. For example, `values=((-np.pi, np.pi), (10, 20), (0, 2 * np.pi))`.
  periodic_boundary : bool
    Whether the neuron encode the value space with the periodic boundary.
  """

  def __init__(self, sigmas, max_ws, min_w=None, encoding_values=None,
               periodic_boundary=False, normalize=True, include_self=True):
    super(DOGDecay, self).__init__()
    self.sigma_p, self.sigma_n = sigmas
    self.max_w_p, self.max_w_n = max_ws
    self.min_w = 0.005 * min(self.max_w_p, self.max_w_n) if min_w is None else min_w
    self.normalize = normalize
    self.include_self = include_self
    self.encoding_values = encoding_values
    self.periodic_boundary = periodic_boundary

  def __call__(self, shape, dtype=None):
    """Build the weights.

    Parameters
    ----------
    shape : tuple of int, list of int, int
      The network shape. Note, this is not the weight shape.
    """
    shape = to_size(shape)

    # value ranges to encode
    if self.encoding_values is None:
      value_ranges = tuple([(0, s) for s in shape])
    elif isinstance(self.encoding_values, (tuple, list)):
      if len(self.encoding_values) == 0:
        raise ValueError
      elif isinstance(self.encoding_values[0], (int, float)):
        assert len(self.encoding_values) == 2
        assert self.encoding_values[0] < self.encoding_values[1]
        value_ranges = tuple([self.encoding_values for _ in shape])
      elif isinstance(self.encoding_values[0], (tuple, list)):
        if len(self.encoding_values) != len(shape):
          raise ValueError(f'The network size has {len(shape)} dimensions, while '
                           f'the encoded values provided only has {len(self.encoding_values)}-D. '
                           f'Error in {str(self)}.')
        for v in self.encoding_values:
          assert isinstance(v[0], (int, float))
          assert len(v) == 2
        value_ranges = tuple(self.encoding_values)
      else:
        raise ValueError(f'Unsupported encoding values: {self.encoding_values}')
    else:
      raise ValueError(f'Unsupported encoding values: {self.encoding_values}')

    # values
    values = [np.linspace(vs[0], vs[1], n + 1)[:n] for vs, n in zip(value_ranges, shape)]
    post_values = np.stack([v.flatten() for v in np.meshgrid(*values)])
    value_sizes = np.array([v[1] - v[0] for v in value_ranges])
    if value_sizes.ndim < post_values.ndim:
      value_sizes = np.expand_dims(value_sizes, axis=tuple([i + 1 for i in range(post_values.ndim - 1)]))
    voxel_ids = np.meshgrid(*[np.arange(s) for s in shape])
    if np.ndim(voxel_ids[0]) > 1:
      voxel_ids = tuple(np.moveaxis(m, 0, 1).flatten() for m in voxel_ids)

    # connectivity matrix
    ndim = len(voxel_ids)
    conn_weights = []
    for v_id in range(voxel_ids[0].shape[0]):
      i_value = []
      for i in range(ndim):
        p_id = voxel_ids[i][v_id]  # position id
        i_value.append(values[i][p_id])
      i_value = np.array(i_value)
      if i_value.ndim < post_values.ndim:
        i_value = np.expand_dims(i_value, axis=tuple([i + 1 for i in range(post_values.ndim - 1)]))
      # distances
      dists = np.abs(i_value - post_values)
      if self.periodic_boundary:
        dists = np.where(dists > value_sizes / 2, value_sizes - dists, dists)
      dists_exp_p = self.max_w_p * np.exp(-(np.linalg.norm(dists, axis=0) / self.sigma_p) ** 2 / 2)
      dists_exp_n = self.max_w_n * np.exp(-(np.linalg.norm(dists, axis=0) / self.sigma_n) ** 2 / 2)
      conn_weights.append(dists_exp_p - dists_exp_n)
    conn_weights = np.stack(conn_weights)
    if not self.include_self:
      np.fill_diagonal(conn_weights, 0.)

    # connectivity weights
    conn_weights = np.where(np.abs(conn_weights) < self.min_w, 0., conn_weights)
    return bm.asarray(conn_weights, dtype=dtype)

  def __repr__(self):
    name = self.__class__.__name__
    bank = ' ' * len(name)
    return (f'{name}(sigmas={(self.sigma_p, self.sigma_n)}, '
            f'max_ws={(self.max_w_p, self.max_w_n)}, min_w={self.min_w}, \n'
            f'{bank}periodic_boundary={self.periodic_boundary}, '
            f'include_self={self.include_self}, '
            f'normalize={self.normalize})')



# -*- coding: utf-8 -*-

from typing import Optional, Union, Callable

import brainpy.math as bm
from brainpy.initialize import Normal, ZeroInit, Initializer, init_param
from brainpy.nn.base import RecurrentNode
from brainpy.nn.datatypes import MultipleData
from brainpy.tools.checking import (check_shape_consistency,
                                    check_float,
                                    check_initializer,
                                    check_string)
from brainpy.types import Tensor

__all__ = [
  'Reservoir',
]


class Reservoir(RecurrentNode):
  r"""Reservoir node, a pool of leaky-integrator neurons
  with random recurrent connections [1]_.

  Parameters
  ----------
  num_unit: int
    The number of reservoir nodes.
  ff_initializer: Initializer
    The initialization method for the feedforward connections.
  rec_initializer: Initializer
    The initialization method for the recurrent connections.
  fb_initializer: optional, Tensor, Initializer
    The initialization method for the feedback connections.
  bias_initializer: optional, Tensor, Initializer
    The initialization method for the bias.
  leaky_rate: float
    A float between 0 and 1.
  activation : str, callable, optional
    Reservoir activation function.
    - If a str, should be a :py:mod:`brainpy.math.activations` function name.
    - If a callable, should be an element-wise operator on tensor.
  activation_type : str
    - If "internal" (default), then leaky integration happens on states transformed
      by the activation function:

    .. math::

        r[n+1] = (1 - \alpha) \cdot r[t] +
        \alpha \cdot f(W_{ff} \cdot u[n] + W_{fb} \cdot b[n] + W_{rec} \cdot r[t])

    - If "external", then leaky integration happens on internal states of
      each neuron, stored in an ``internal_state`` parameter (:math:`x` in
      the equation below).
      A neuron internal state is the value of its state before applying
      the activation function :math:`f`:

      .. math::

          x[n+1] &= (1 - \alpha) \cdot x[t] +
          \alpha \cdot f(W_{ff} \cdot u[n] + W_{rec} \cdot r[t] + W_{fb} \cdot b[n]) \\
          r[n+1] &= f(x[n+1])
  ff_connectivity : float, optional
    Connectivity of input neurons, i.e. ratio of input neurons connected
    to reservoir neurons. Must be in [0, 1], by default 0.1
  rec_connectivity : float, optional
    Connectivity of recurrent weights matrix, i.e. ratio of reservoir
    neurons connected to other reservoir neurons, including themselves.
    Must be in [0, 1], by default 0.1
  fb_connectivity : float, optional
    Connectivity of feedback neurons, i.e. ratio of feedabck neurons
    connected to reservoir neurons. Must be in [0, 1], by default 0.1
  conn_type: str
    The connectivity type, can be "dense" or "sparse".
  spectral_radius : float, optional
    Spectral radius of recurrent weight matrix, by default None
  noise_rec : float, optional
    Gain of noise applied to reservoir internal states, by default 0.0
  noise_in : float, optional
    Gain of noise applied to feedforward signals, by default 0.0
  noise_fb : float, optional
    Gain of noise applied to feedback signals, by default 0.0
  noise_type : optional, str, callable
    Distribution of noise. Must be a random variable generator
    distribution (see :py:class:`brainpy.math.random.RandomState`),
    by default "normal". 
  seed: optional, int
    The seed for random sampling in this node.

  References
  ----------
  .. [1] Lukoševičius, Mantas. "A practical guide to applying echo state networks."
         Neural networks: Tricks of the trade. Springer, Berlin, Heidelberg, 2012. 659-686.
  """
  data_pass = MultipleData('sequence')

  def __init__(
      self,
      num_unit: int,
      leaky_rate: float = 0.3,
      activation: Union[str, Callable] = 'tanh',
      activation_type: str = 'internal',
      ff_initializer: Union[Initializer, Callable, Tensor] = Normal(scale=0.1),
      rec_initializer: Union[Initializer, Callable, Tensor] = Normal(scale=0.1),
      fb_initializer: Optional[Union[Initializer, Callable, Tensor]] = Normal(scale=0.1),
      bias_initializer: Optional[Union[Initializer, Callable, Tensor]] = ZeroInit(),
      ff_connectivity: float = 0.1,
      rec_connectivity: float = 0.1,
      fb_connectivity: float = 0.1,
      conn_type='dense',
      spectral_radius: Optional[float] = None,
      noise_ff: float = 0.,
      noise_rec: float = 0.,
      noise_fb: float = 0.,
      noise_type: str = 'normal',
      seed: Optional[int] = None,
      trainable: bool = False,
      **kwargs
  ):
    super(Reservoir, self).__init__(trainable=trainable, **kwargs)

    # parameters
    self.num_unit = num_unit
    assert num_unit > 0, f'Must be a positive integer, but we got {num_unit}'
    self.leaky_rate = leaky_rate
    check_float(leaky_rate, 'leaky_rate', 0., 1.)
    self.activation = bm.activations.get(activation)
    self.activation_type = activation_type
    check_string(activation_type, 'activation_type', ['internal', 'external'])
    self.rng = bm.random.RandomState(seed)
    check_float(spectral_radius, 'spectral_radius', allow_none=True)
    self.spectral_radius = spectral_radius

    # initializations
    check_initializer(ff_initializer, 'ff_initializer', allow_none=False)
    check_initializer(rec_initializer, 'rec_initializer', allow_none=False)
    check_initializer(fb_initializer, 'fb_initializer', allow_none=True)
    check_initializer(bias_initializer, 'bias_initializer', allow_none=True)
    self.ff_initializer = ff_initializer
    self.fb_initializer = fb_initializer
    self.rec_initializer = rec_initializer
    self.bias_initializer = bias_initializer

    # connectivity
    check_float(ff_connectivity, 'ff_connectivity', 0., 1.)
    check_float(rec_connectivity, 'rec_connectivity', 0., 1.)
    check_float(fb_connectivity, 'fb_connectivity', 0., 1.)
    self.ff_connectivity = ff_connectivity
    self.rec_connectivity = rec_connectivity
    self.fb_connectivity = fb_connectivity
    check_string(conn_type, 'conn_type', ['dense', 'sparse'])
    self.conn_type = conn_type

    # noises
    check_float(noise_ff, 'noise_ff')
    check_float(noise_fb, 'noise_fb')
    check_float(noise_rec, 'noise_rec')
    self.noise_ff = noise_ff
    self.noise_fb = noise_fb
    self.noise_rec = noise_rec
    self.noise_type = noise_type
    check_string(noise_type, 'noise_type', ['normal', 'uniform'])

  def init_ff_conn(self):
    """Initialize feedforward connections, weights, and variables."""
    unique_shape, free_shapes = check_shape_consistency(self.feedforward_shapes, -1, True)
    self.set_output_shape(unique_shape + (self.num_unit,))

    # initialize feedforward weights
    weight_shape = (sum(free_shapes), self.num_unit)
    self.Wff = init_param(self.ff_initializer, weight_shape)
    if self.ff_connectivity < 1.:
      conn_mat = self.rng.random(weight_shape) > self.ff_connectivity
      self.Wff[conn_mat] = 0.
    if self.conn_type == 'sparse' and self.ff_connectivity < 1.:
      self.ff_pres, self.ff_posts = bm.where(bm.logical_not(conn_mat))
      self.Wff = self.Wff[self.ff_pres, self.ff_posts]
    if self.trainable:
      self.Wff = bm.TrainVar(self.Wff)

    # initialize recurrent weights
    recurrent_shape = (self.num_unit, self.num_unit)
    self.Wrec = init_param(self.rec_initializer, recurrent_shape)
    if self.rec_connectivity < 1.:
      conn_mat = self.rng.random(recurrent_shape) > self.rec_connectivity
      self.Wrec[conn_mat] = 0.
    if self.spectral_radius is not None:
      current_sr = max(abs(bm.linalg.eig(self.Wrec)[0]))
      self.Wrec *= self.spectral_radius / current_sr
    if self.conn_type == 'sparse' and self.rec_connectivity < 1.:
      self.rec_pres, self.rec_posts = bm.where(bm.logical_not(conn_mat))
      self.Wrec = self.Wrec[self.rec_pres, self.rec_posts]
    self.bias = init_param(self.bias_initializer, (self.num_unit,))
    if self.trainable:
      self.Wrec = bm.TrainVar(self.Wrec)
      self.bias = None if (self.bias is None) else bm.TrainVar(self.bias)

    # initialize feedback weights
    self.Wfb = None

  def init_state(self, num_batch=1):
    # initialize internal state
    return bm.zeros((num_batch, self.num_unit), dtype=bm.float_)

  def init_fb_conn(self):
    """Initialize feedback connections, weights, and variables."""
    if self.feedback_shapes is not None:
      unique_shape, free_shapes = check_shape_consistency(self.feedback_shapes, -1, True)
      fb_shape = (sum(free_shapes), self.num_unit)
      self.Wfb = init_param(self.fb_initializer, fb_shape)
      if self.fb_connectivity < 1.:
        conn_mat = self.rng.random(fb_shape) > self.fb_connectivity
        self.Wfb[conn_mat] = 0.
      if self.conn_type == 'sparse' and self.fb_connectivity < 1.:
        self.fb_pres, self.fb_posts = bm.where(bm.logical_not(conn_mat))
        self.Wfb = self.Wfb[self.fb_pres, self.fb_posts]
      if self.trainable:
        self.Wfb = bm.TrainVar(self.Wfb)

  def forward(self, ff, fb=None, **shared_kwargs):
    """Feedforward output."""
    # inputs
    x = bm.concatenate(ff, axis=-1)
    if self.noise_ff > 0: x += self.noise_ff * self.rng.uniform(-1, 1, x.shape)
    if self.conn_type == 'sparse' and self.ff_connectivity < 1.:
      hidden = bm.sparse_matmul(x, [self.Wff, (self.ff_pres, self.ff_posts)], self.num_unit)
    else:
      hidden = bm.dot(x, self.Wff)
    # feedback
    if self.Wfb is not None:
      assert fb is not None, 'Should provide feedback signals, but we got None.'
      fb = bm.concatenate(fb, axis=-1)
      if self.noise_fb: fb += self.noise_fb * self.rng.uniform(-1, 1, fb.shape)
      if self.conn_type == 'sparse' and self.fb_connectivity < 1.:
        hidden += bm.sparse_matmul(fb, [self.Wfb, (self.fb_pres, self.fb_posts)], self.num_unit)
      else:
        hidden += bm.dot(fb, self.Wfb)
    # recurrent
    if self.conn_type == 'sparse' and self.rec_connectivity < 1.:
      hidden += bm.sparse_matmul(self.state, [self.Wrec, (self.rec_pres, self.rec_posts)], self.num_unit)
    else:
      hidden += bm.dot(self.state, self.Wrec)
    if self.activation_type == 'internal':
      hidden = self.activation(hidden)
    if self.noise_rec > 0.: hidden += self.noise_rec * self.rng.uniform(-1, -1, self.state.shape)
    # new state/output
    state = (1 - self.leaky_rate) * self.state + self.leaky_rate * hidden
    if self.activation_type == 'external':
      state = self.activation(state)
    self.state.value = state
    return state



# -*- coding: utf-8 -*-

"""
This module aims to provide commonly used analysis methods for simulated neuronal data.
You can access them through ``brainpy.measure.XXX``.
"""

from .correlation import *
from .firings import *




# -*- coding: utf-8 -*-


__all__ = [
  'visualize'
]


class visualize(object):
  @staticmethod
  def get_figure(row_num,
                 col_num,
                 row_len=3,
                 col_len=6):
    from .figures import get_figure
    return get_figure(row_num, col_num, row_len, col_len)

  @staticmethod
  def line_plot(ts,
                val_matrix,
                plot_ids=None,
                ax=None,
                xlim=None,
                ylim=None,
                xlabel='Time (ms)',
                ylabel=None,
                legend=None,
                title=None,
                show=False,
                **kwargs):
    from .plots import line_plot
    line_plot(ts, val_matrix, plot_ids=plot_ids, ax=ax, xlim=xlim, ylim=ylim,
              xlabel=xlabel, ylabel=ylabel, legend=legend, title=title, show=show, **kwargs)

  @staticmethod
  def raster_plot(ts,
                  sp_matrix,
                  ax=None,
                  marker='.',
                  markersize=2,
                  color='k',
                  xlabel='Time (ms)',
                  ylabel='Neuron index',
                  xlim=None,
                  ylim=None,
                  title=None,
                  show=False,
                  **kwargs):
    from .plots import raster_plot
    raster_plot(ts, sp_matrix, ax=ax, marker=marker, markersize=markersize, color=color,
                xlabel=xlabel, ylabel=ylabel, xlim=xlim, ylim=ylim, title=title, show=show, **kwargs)

  @staticmethod
  def animate_1D(dynamical_vars,
                 static_vars=(),
                 dt=None,
                 xlim=None,
                 ylim=None,
                 xlabel=None,
                 ylabel=None,
                 frame_delay=50.,
                 frame_step=1,
                 title_size=10,
                 figsize=None,
                 gif_dpi=None,
                 video_fps=None,
                 save_path=None,
                 show=True,
                 **kwargs):
    from .plots import animate_1D
    return animate_1D(dynamical_vars, static_vars=static_vars, dt=dt, xlim=xlim, ylim=ylim,
                      xlabel=xlabel, ylabel=ylabel, frame_delay=frame_delay, frame_step=frame_step,
                      title_size=title_size, figsize=figsize, gif_dpi=gif_dpi, video_fps=video_fps,
                      save_path=save_path, show=show, **kwargs)

  @staticmethod
  def animate_2D(values,
                 net_size,
                 dt=None,
                 val_min=None,
                 val_max=None,
                 cmap=None,
                 frame_delay=10,
                 frame_step=1,
                 title_size=10,
                 figsize=None,
                 gif_dpi=None,
                 video_fps=None,
                 save_path=None,
                 show=True):
    from .plots import animate_2D
    return animate_2D(values, net_size, dt=dt, val_min=val_min, val_max=val_max, cmap=cmap,
                      frame_delay=frame_delay, frame_step=frame_step, title_size=title_size,
                      figsize=figsize, gif_dpi=gif_dpi, video_fps=video_fps, save_path=save_path, show=show)

  @staticmethod
  def plot_style1(fontsize=22,
                  axes_edgecolor='black',
                  figsize='5,4',
                  lw=1):
    from .styles import plot_style1
    plot_style1(fontsize=fontsize, axes_edgecolor=axes_edgecolor, figsize=figsize, lw=lw)



# -*- coding: utf-8 -*-

"""
The ``base`` module for whole BrainPy ecosystem.

- This module provides the most fundamental class ``Base``,
  and its associated helper class ``Collector`` and ``ArrayCollector``.
- For each instance of "Base" class, users can retrieve all
  the variables (or trainable variables), integrators, and nodes.
- This module also provides a ``Function`` class to wrap user-defined
  functions. In each function, maybe several nodes are used, and
  users can initialize a ``Function`` by providing the nodes used
  in the function. Unfortunately, ``Function`` class does not have
  the ability to gather nodes automatically.
- This module provides ``io`` helper functions to help users save/load
  model states, or share user's customized model with others.
- This module provides ``naming`` tools to guarantee the unique nameing
  for each Base object.

Details please see the following.
"""

from brainpy.base.base import *
from brainpy.base.collector import *
from brainpy.base.function import *
from brainpy.base.io import *
from brainpy.base.naming import *

